# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
High level API for the polyply coordinate generator
"""
import sys
import numpy as np
import networkx as nx
import vermouth.forcefield
from vermouth.file_writer import DeferredFileWriter
from vermouth.log_helpers import StyleAdapter, get_logger
from .generate_templates import GenerateTemplates
from .backmap import Backmap
from .topology import Topology
from .build_system import BuildSystem
from .annotate_ligands import AnnotateLigands, parse_residue_spec, _find_nodes
from .build_file_parser import read_build_file
from .check_residue_equivalence import check_residue_equivalence

LOGGER = StyleAdapter(get_logger(__name__))

def find_starting_node_from_spec(topology, start_nodes):
    """
    Given a definition of one or multiple nodes in
    `start_nodes` find the corresponding molecule index
    and node key in `topology`. The format of start nodes
    is equivalent to that for annotating the residue and
    is described in detail in the function
    :meth:`polyply.src.annotate_ligands.annotate_molecules.parse_residue_spec`.

    Parameters:
    -----------
    topology: :class:`polyply.src.topology.Topology`
    start_nodes: list[str]
        a list of residue-specs

    Returns:
    --------
    dict
        a dict of molecule_idx as key and node key as value
    """
    start_dict = {mol_idx: None for mol_idx in range(len(topology.molecules))}
    for start in start_nodes:
        res_spec = parse_residue_spec(start)
        if 'mol_idx' in res_spec:
            mol_idx = res_spec['mol_idx']
            node = list(_find_nodes(topology.molecules[mol_idx], res_spec))[0]
            start_dict[mol_idx] = node
            topology.molecules[mol_idx].root = node
        else:
            for idx, molecule in enumerate(topology.molecules):
                if molecule.mol_name == res_spec['molname']:
                    node = list(_find_nodes(molecule, res_spec))[0]
                    start_dict[idx] = node
    return start_dict

def _initialize_cylces(topology, cycles, tolerance):
    for mol_name in cycles:
        for mol_idx in topology.mol_idx_by_name[mol_name]:
            # initalize as dfs tree
            molecule = topology.molecules[mol_idx]
            cycles = nx.cycle_basis(molecule)
            if len(cycles) > 1:
                raise IOError("More than one cycle is not allowed.")
            molecule.dfs=True
            nodes = (list(molecule.search_tree.edges)[0][0],
                     list(molecule.search_tree.edges)[-1][1])
            topology.distance_restraints[(mol_name, mol_idx)][nodes] = (0.0, tolerance)

def _check_molecules(molecules):
    """
    Helper method which raises an IOError
    if any molecule in `molecules` is
    disconnected.
    """
    # check if molecules are all connected
    for molecule in molecules:
        if not nx.is_connected(molecule):
            msg = ('\n Molecule {} consistes of two disconnected parts. '
                   'Make sure all atoms/particles in a molecule are '
                   'connected by bonds, constraints or virual-sites')
            raise IOError(msg.format(molecule.name))

def gen_coords(args):
    # Read in the topology
    LOGGER.info("reading topology",  type="step")
    topology = Topology.from_gmx_topfile(name=args.name, path=args.toppath)

    LOGGER.info("processing topology",  type="step")
    topology.preprocess()
    _check_molecules(topology.molecules)

    if args.split:
        LOGGER.info("splitting residues",  type="step")
        for molecule in topology.molecules:
            molecule.split_residue(args.split)

    # read in coordinates if there are any
    if args.coordpath:
        LOGGER.info("loading coordinates",  type="step")
        topology.add_positions_from_file(args.coordpath, args.build_res)

    # load in built file
    if args.build:
        LOGGER.info("reading build file",  type="step")
        with open(args.build) as build_file:
            lines = build_file.readlines()
            read_build_file(lines, topology.molecules, topology)

    # collect all starting points for the molecules
    start_dict = find_starting_node_from_spec(topology, args.start)

    # handle grid input
    if args.grid:
        LOGGER.info("loading grid",  type="step")
        args.grid = np.loadtxt(args.grid)

    # do a sanity check
    LOGGER.info("checking residue integrity",  type="step")
    check_residue_equivalence(topology)
    # Build polymer structure
    LOGGER.info("generating templates",  type="step")
    GenerateTemplates(topology=topology, max_opt=10).run_system(topology)
    LOGGER.info("annotating ligands",  type="step")
    ligand_annotator = AnnotateLigands(topology, args.ligands)
    ligand_annotator.run_system(topology)
    LOGGER.info("generating system coordinates",  type="step")
    _initialize_cylces(topology, args.cycles, args.cycle_tol)
    BuildSystem(topology,
                start_dict=start_dict,
                density=args.density,
                max_force=args.max_force,
                grid_spacing=args.grid_spacing,
                maxiter=args.maxiter,
                box=args.box,
                step_fudge=args.step_fudge,
                ignore=args.ignore,
                grid=args.grid,
                cycles=args.cycles,
                nrewind=args.nrewind).run_system(topology.molecules)
    ligand_annotator.split_ligands()
    LOGGER.info("backmapping to target resolution",  type="step")
    Backmap(fudge_coords=args.bfudge).run_system(topology)
    # Write output
    LOGGER.info("writing output",  type="step")
    command = ' '.join(sys.argv)
    system = topology.convert_to_vermouth_system()
    vermouth.gmx.gro.write_gro(system, args.outpath, precision=7,
                               title=command, box=topology.box)
    DeferredFileWriter().write()



# Copyright 2022 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict
from functools import partial
import json
import networkx as nx
from networkx.readwrite import json_graph
from vermouth.parser_utils import split_comments
from vermouth.log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))

ONE_LETTER_DNA = {"A": "DA",
                  "C": "DC",
                  "G": "DG",
                  "T": "DT"}

ONE_LETTER_RNA = {"A": "A",
                  "C": "C",
                  "G": "G",
                  "T": "U"}

class FileFormatError(Exception):
    """Raised when a parser fails due to invalid file format."""

def _monomers_to_linear_nx_graph(monomers):
    """
    Take a list of monomers and generate a linear
    graph from them setting both the resid and
    resname attribute.
    """
    seq_graph = nx.Graph()
    mon_range = range(0, len(monomers))
    seq_graph.add_edges_from(zip(mon_range[:-1], mon_range[1:]))
    nx.set_node_attributes(seq_graph, dict(zip(seq_graph.nodes, monomers)), "resname")
    nx.set_node_attributes(seq_graph, {node: node+1 for node in seq_graph.nodes}, "resid")
    return seq_graph

def _parse_plain_delimited(filepath, delimiter=" "):
    """
    Parse a plain delimited text file. The delimiter can
    be any special character. Note the current function
    has only been tested for space delimeted txt files
    so far. Use other delimiters at own risk or make
    sure the parsing behaviour is as expected.

    Parameters
    ----------
    filepath: str or path
    delimeter: str
        any string character that delimits the file

    Returns
    --------
    :class:`nx.Graph`
        A plain graph of the molecular sequence with
        node attributes resname and resid
    """
    with open(filepath) as file_:
        lines = file_.readlines()

    monomers = []
    for line in lines:
        for resname in line.strip().split(delimiter):
            monomers.append(resname.strip())
    seq_graph =  _monomers_to_linear_nx_graph(monomers)
    return seq_graph

parse_txt = _parse_plain_delimited

def _parse_plain(lines, DNA=False, RNA=False):
    """
    Parse a plain one letter sequence block either for DNA, RNA,
    or amino-acids. Lines can be a list of strings or a string.
    This function also sets the appropiate defaults for the termini
    of DNA and RNA.

    For the format see here:

    https://www.animalgenome.org/bioinfo/resources/manuals/seqformats#plain

    Parameters
    ----------
    lines: `abc.iteratable`
        list of strings matching one letter code DNA, RNA, or AAs
    DNA: bool
        if the sequence matches DNA
    RNA: bool
        if the sequence matches RNA

    Returns
    -------
    `:class:nx.Graph`
        A plain graph of the molecular sequence with
        node attributes resname and resid

    Raises
    ------
    IOError
        If the one letter residue codes don't match any
        kown nucleobase
    """
    monomers = []
    for line in lines:
        for token in line.strip():
            if token in ONE_LETTER_DNA and DNA:
                resname = ONE_LETTER_DNA[token]
            elif token in ONE_LETTER_RNA and RNA:
                resname = ONE_LETTER_RNA[token]
            else:
                msg = f"Cannot find one letter residue match for {token}"
                raise IOError(msg)

            monomers.append(resname)

    # make sure to set the defaults for the DNA and RNA terminals
    monomers[0] = monomers[0] + "5"
    monomers[-1] = monomers[-1] + "3"

    seq_graph =  _monomers_to_linear_nx_graph(monomers)
    return seq_graph

def _identify_nucleotypes(comments):
    """
    From a comment found in the ig or fasta file, identify if
    the sequence is RNA or DNA sequence by checking if these
    keywords are in the comment lines. Raise an error if
    none or conflicting information are found.

    Parameters
    ----------
    comments: abc.itertable
        list of comment lines

    Returns
    -------
    bool, bool
        is it DNA, RNA

    Raises
    ------
    FileFormatError
        neither RNA nor DNA keywords are found
        both RNA and DNA are found
    """
    RNA = False
    DNA = False
    for comment in comments:
        if "DNA" in comment:
            DNA = True

        if "RNA" in comment:
            RNA = True

    if RNA and DNA:
        raise FileFormatError("Found both RNA and DNA keyword in comment. Choose one.")

    if not RNA and not DNA:
        raise FileFormatError("Cannot identify if sequence is RNA or DNA from comment.")

    return DNA, RNA

def parse_ig(filepath):
    """
    Read ig sequence in DNA/RNA formatted file.

    The parser automatically translates the one letter code to the
    double letter nucleobase resnames, sets special residue names
    for the 5' and 3' end and identifies if it is RNA or DNA. This is
    done by eading the comments and looking for either term being
    present in the comments. In addition if a circular sequence is given
    the linktype circle attribute is set, which is picked up by the
    :class:`polyply.ApplyLinks` processor for correctly making
    circular itp files.

    See following link for format:

    https://www.animalgenome.org/bioinfo/resources/manuals/seqformats#IG

    Parameters
    ----------
    filepath: :class:`pathlib.Path`


    Returns
    --------
    :class:`nx.Graph`
        A plain graph of the molecular sequence with
        node attributes resname and resid
    """
    with open(filepath) as file_:
        lines = file_.readlines()

    clean_lines = []
    comments = []
    idx = 0
    for idx, line in enumerate(lines):
        clean_line, comment = split_comments(line)
        comments.append(comment)
        if clean_line:
            if clean_line[-1] == '1' or clean_line[-1] == '2':
                ter_char = clean_line[-1]
                clean_line = clean_line[:-1]
                clean_lines.append(clean_line)
                break
            else:
                clean_lines.append(clean_line)
    else:
        msg = "The sequence is not complete, it does not end with 1 or 2."
        raise FileFormatError(msg)

    DNA, RNA = _identify_nucleotypes(comments)
    seq_graph = _parse_plain(clean_lines[1:], DNA=DNA, RNA=RNA)

    if ter_char == '2':
        nnodes = len(seq_graph.nodes)
        seq_graph.add_edge(0, nnodes-1)
        seq_graph.edges[(0, nnodes-1)]["linktype"] = "circle"
        seq_graph.nodes[0]["resname"] = seq_graph.nodes[0]["resname"][:-1]
        seq_graph.nodes[nnodes-1]["resname"] = seq_graph.nodes[nnodes-1]["resname"][:-1]

    if idx < len(lines) - 1:
        LOGGER.warning("There may be more than one sequence in the file. We will only use the first one.")

    return seq_graph

def parse_fasta(filepath):
    """
    Read fasta sequence of DNA/RNA.

    The parser automatically translates the one letter code to the
    double letter nucleobase resnames, sets special residue names
    for the 5' and 3' end and identifies if it is RNA or DNA. This is
    done by eading the comments and looking for either term being
    present in the comments.


    See following link for format:
    https://www.animalgenome.org/bioinfo/resources/manuals/seqformats#FASTA

    Parameters
    ----------
    filepath: str or path
        path to file

    Returns
    --------
    :class:`nx.Graph`
        A plain graph of the molecular sequence with
        node attributes resname and resid
    """
    with open(filepath) as file_:
        lines = file_.readlines()

    clean_lines = []
    # first line must be a comment line
    DNA, RNA =_identify_nucleotypes([lines[0]])

    for line in lines[1:]:
        if '>' in line:
            LOGGER.warning("Found more than 1 sequence. Will only use the first one.")
            break

        clean_lines.append(line)

    seq_graph = _parse_plain(clean_lines, RNA=RNA, DNA=DNA)
    return seq_graph

def parse_json(filepath):
    """
    Read in json file that specifies a molecular
    sequence at the residue level including connectivity.

    Parameters
    ----------
    filepath: str or path
        path to file

    Returns
    --------
    :class:`nx.Graph`
        A plain graph of the molecular sequence with
        node attributes resname and resid

    """
    with open(filepath) as file_:
        data = json.load(file_)

    init_json_graph = nx.Graph(json_graph.node_link_graph(data))
    # the nodes in the inital graph are not ordered, when no resid
    # is given this can create issues. So we reorder the node based
    # on the node key, which HAS to be numerical.
    seq_graph = nx.Graph(node_dict_factory=OrderedDict)
    nodes = list(init_json_graph.nodes(data=True))
    nodes.sort()
    seq_graph.add_nodes_from(nodes)
    seq_graph.add_edges_from(init_json_graph.edges(data=True))
    return seq_graph



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that build files are properly read.
"""
import textwrap
import pytest
import numpy as np
import networkx as nx
import vermouth.forcefield
import vermouth.molecule
import polyply
from polyply.src.meta_molecule import Monomer, MetaMolecule
from polyply.src.topology import Topology

@pytest.mark.parametrize('tokens, _type, expected', (
   # for cylinder
   (["PEO", "63", "154", "in", "1", "2", "3", "6", "7"],
    "cylinder",
    {"resname": "PEO", "start": 63, "stop": 154, "parameters":["in", np.array([1.0, 2.0, 3.0]), 6.0, 7.0, "cylinder"]})
   # for recangle
   ,(["PEO", "0", "10", "out", "11", "12", "13", "1", "2", "3"],
    "rectangle",
    {"resname": "PEO", "start": 0, "stop": 10, "parameters":["out", np.array([11.0, 12.0, 13.0]), 1.0, 2.0, 3.0, "rectangle"]})
   # for sphere
   ,(["PEO", "0", "10", "in", "11", "12", "13", "5"],
    "sphere",
    {"resname": "PEO", "start": 0, "stop": 10, "parameters":["in", np.array([11.0, 12.0, 13.0]), 5.0, "sphere"]})
   ))
def test_base_parser_geometry(tokens, _type, expected):
    result = polyply.src.build_file_parser.BuildDirector._base_parser_geometry(tokens, _type)
    assert result.keys() == expected.keys()
    for key in result:
        if key != "parameters":
            assert result[key] == expected[key]
        else:
            assert result[key][0] == expected[key][0]
            assert all(result[key][1] == expected[key][1])
            for result_param, expected_param in zip(result[key][2:], expected[key][2:]):
                assert result_param == expected_param

@pytest.fixture
def test_molecule():
    # dummy vermouth force-field
    force_field = vermouth.forcefield.ForceField(name='test_ff')
    # monomers used in the meta-molecule
    ALA = Monomer(resname="ALA", n_blocks=4)
    GLU = Monomer(resname="GLU", n_blocks=2)
    THR = Monomer(resname="THR", n_blocks=3)
    # two meta-molecules
    mol = MetaMolecule.from_monomer_seq_linear(force_field,
                                               [ALA, GLU, THR],
                                               "AA")
    return mol

@pytest.mark.parametrize('_type, option, expected', (
   # tag all ALA nodes
   ("cylinder",
    {"resname": "ALA", "start": 1, "stop": 4, "parameters":["in", np.array([5.0, 5.0, 5.0]), 5.0, 5.0]},
    [0, 1, 2, 3]),
   # tag all nodes 1-4 that have name ALA, GLU, THR should not get tagged
   ("cylinder",
    {"resname": "ALA", "start": 1, "stop": 9, "parameters":["in", np.array([5.0, 5.0, 5.0]), 5.0, 5.0]},
   [0, 1, 2, 3]),
   # tag only GLU nodes
   ("sphere",
    {"resname": "GLU", "start": 5, "stop": 6, "parameters":["in", np.array([10.0, 10.0, 10.0]), 5.0]},
   [4, 5]),
   # tag no nodes based on the resid rensame correspondance
   ("sphere",
    {"resname": "GLU", "start": 1, "stop": 4, "parameters":["in", np.array([10.0, 10.0, 10.0]), 5.0]},
   [])
   # tag no nodes based on no resname in molecule
   ,("sphere",
    {"resname": "PPI", "start": 1, "stop": 4, "parameters":["in", np.array([10.0, 10.0, 10.0]), 5.0]},
   [])
   ))
def test_tag_nodes(test_molecule, _type, option, expected):
    polyply.src.build_file_parser.BuildDirector._tag_nodes(test_molecule, _type, option)
    for node in test_molecule.nodes:
        if "restraints" in test_molecule.nodes[node]:
           assert node in expected

@pytest.fixture()
def test_system():
  """
  Create a dummy test system with three types of molecules AA, BB and
  NA. NA is the molecule to be used a ligand. AA and BB are composed
  of different residues.
  """
  # dummy vermouth force-field
  force_field = vermouth.forcefield.ForceField(name='test_ff')
  # monomers used in the meta-molecule
  ALA = Monomer(resname="ALA", n_blocks=4)
  GLU = Monomer(resname="GLU", n_blocks=2)
  THR = Monomer(resname="THR", n_blocks=3)
  # two meta-molecules
  meta_mol_A = MetaMolecule.from_monomer_seq_linear(force_field,
                                                    [ALA, GLU, THR],
                                                    "AA")
  meta_mol_B = MetaMolecule.from_monomer_seq_linear(force_field,
                                                    [GLU, ALA, THR],
                                                    "BB")
  NA = MetaMolecule()
  NA.add_monomer(current=0, resname="NA", connections=[])
  molecules = [meta_mol_A, meta_mol_A.copy(),
               meta_mol_B.copy(), NA, NA.copy(),
               NA.copy(), NA.copy()]
  top = Topology(force_field=force_field)
  top.molecules = molecules
  top.mol_idx_by_name = {"AA":[0, 1], "BB": [2], "NA":[3, 4, 5, 6]}
  return top

@pytest.mark.parametrize('line, expected', (
   # simple example
   ("ALA 0 4 5.0 3.0 2.0 50.0",
    {"resname": "ALA", "start": 0, "stop": 4, "parameters": [np.array([5.0, 3.0, 2.0]), 50.]})
   # other example
   ,("GLU 5 6 4.0 4.0 1.0 25.0",
    {"resname": "GLU", "start": 5, "stop": 6, "parameters": [np.array([4.0, 4.0, 1.0]), 25.0]})
   ))
def test_rw_restriction_parsing(test_system, line, expected):
    processor = polyply.src.build_file_parser.BuildDirector([], test_system)
    processor.current_molidxs = [1]
    processor.current_molname = "AA"
    processor._rw_restriction(line)
    result =  processor.rw_options[("AA", 1)]
    assert result.keys() == expected.keys()
    for key in processor.rw_options[("AA", 1)]:
        if key != "parameters":
            assert result[key] == expected[key]
        else:
            assert all(result[key][0] == expected[key][0])
            assert result[key][1] == expected[key][1]

@pytest.mark.parametrize('line, key, expected', (
   # simple example
   ("1 8 3.0",
    (1, 8),
    (3.0, 0.0)),
   ("1 8 3.0 1.0",
    (1, 8),
    (3.0, 1.0)),
   ))
def test_distance_restraints(test_system, line, key, expected):
    processor = polyply.src.build_file_parser.BuildDirector([], test_system)
    processor.current_molidxs = [0]
    processor.current_molname = "AA"
    processor._distance_restraints(line)
    result = test_system.distance_restraints[("AA", 0)]
    result[key] == expected


@pytest.mark.parametrize('line', (
   # basic test
   ("1 50 3.0",
    "50 1 3.0")))
def test_distance_restraints_error(test_system, line):
    processor = polyply.src.build_file_parser.BuildDirector([], test_system)
    processor.current_molidxs = [0]
    processor.current_molname = "AA"
    with pytest.raises(IOError):
        processor._distance_restraints(line)


@pytest.mark.parametrize('lines, tagged_mols, tagged_nodes', (
   # basic test
   ("""
   [ molecule ]
   ; name from to
   AA    0  2
   ;
   [ cylinder ]
   ; resname start stop  inside-out  x  y  z  r   z
   ALA   2    4  in  5  5  5  5  5
   """,
   [0, 1],
   [0, 1, 2, 3]),
   # test that a different molecule is tagged
   ("""
   [ molecule ]
   BB  2  3
   [ rectangle ]
   ; resname start stop  inside-out  x  y  z a b c
   ALA   3    6  in  5  5  5  5  5 5
   """,
   [2],
   [2, 3, 4, 5]),
   # test nothing is tagged based on the molname
   ("""
   [ molecule ]
   CC 1 6
   [ sphere ]
   ; resname start stop  inside-out  x  y  z r
   ALA   2    4  in  5  5  5  5
   """,
   [],
   []),
   ))
def test_parser(test_system, lines, tagged_mols, tagged_nodes):
   lines = textwrap.dedent(lines).splitlines()
   ff = vermouth.forcefield.ForceField(name='test_ff')
   top = Topology(ff)
   polyply.src.build_file_parser.read_build_file(lines,
                                                 test_system.molecules,
                                                 top)
   for idx, mol in enumerate(test_system.molecules):
       for node in mol.nodes:
           if "restraints" in mol.nodes[node]:
               assert node in tagged_nodes
               assert idx in tagged_mols


@pytest.mark.parametrize('lines, expected', (
   # basic test
   ("""
   [ molecule ]
   ; name from to
   AA    0  2
   ;
   [ persistence_length ]
   ; model  lp start stop
     WCM   4.0  0    8
   """,
   [["WCM", 4.0, 0, 8, np.array([0, 1])]]),
   # test two batches are recognized
   ("""
   [ molecule ]
   ; name from to
   AA    0  2
   ;
   [ persistence_length ]
   ; model  lp start stop
   WCM   4.0  0    8
   [ molecule ]
   ; name from to
   BB 2  4
   [ persistence_length ]
   WCM   2.0   0  8
   """,
   [["WCM", 4.0, 0, 8, np.array([0, 1])],
    ["WCM", 2.0, 0, 8, np.array([2, 3])]]
   )))
def test_persistence_parsers(test_system, lines, expected):
   lines = textwrap.dedent(lines).splitlines()
   ff = vermouth.forcefield.ForceField(name='test_ff')
   top = Topology(ff)
   polyply.src.build_file_parser.read_build_file(lines,
                                                 test_system.molecules,
                                                 top)
   for ref, new in zip(expected, top.persistences):
        print(ref, new)
        for info_ref, info_new in zip(ref[:-1], new[:-1]):
            assert info_ref == info_new
        assert all(ref[-1] == new[-1])



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that build files are properly read.
"""

import logging
import pytest
import textwrap
import math
import numpy as np
import networkx as nx
from vermouth.forcefield import ForceField
import polyply
from polyply.src.topology import Topology
from polyply.src.top_parser import read_topology
from polyply.src.nonbond_engine import NonBondEngine
from polyply.src.build_file_parser import PersistenceSpecs
from polyply.src.persistence import sample_end_to_end_distances


@pytest.fixture
def topology():
    top_lines ="""
    [ defaults ]
    1   1   no   1.0     1.0
    [ atomtypes ]
    N0 72.0 0.000 A 0.0 0.0
    [ nonbond_params ]
    N0   N0   1 4.700000e-01    3.700000e+00
    [ moleculetype ]
    testA 1
    [ atoms ]
    1    N0   1   GLY    BB   1 0.00     45
    2    N0   2   GLY    SC1  1 0.00     45
    3    N0   3   GLY    SC2  1 0.00     45
    4    N0   4   GLU    BB   2 0.00     45
    5    N0   5   GLU    SC1  2 0.00     45
    6    N0   6   GLU    SC2  2 0.00     45
    7    N0   7   GLU    SC2  2 0.00     45
    8    N0   8   GLU    SC2  2 0.00     45
    9    N0   9   GLU    SC2  2 0.00     45
   10    N0  10   GLU    SC2  2 0.00     45
   11    N0  11   GLU    SC2  2 0.00     45
   12    N0  12   GLU    SC2  2 0.00     45
   13    N0  13   GLU    SC2  2 0.00     45
   14    N0  14   GLU    SC2  2 0.00     45
   15    N0  15   GLU    SC2  2 0.00     45
   16    N0  16   GLU    SC2  2 0.00     45
   17    N0  17   GLU    SC2  2 0.00     45
   18    N0  18   GLU    SC2  2 0.00     45
   19    N0  19   GLU    SC2  2 0.00     45
   20    N0  20   GLU    SC2  2 0.00     45
   21    N0  21   GLU    SC2  2 0.00     45
   22    N0  22   GLU    SC2  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    6     7    1  0.47 2000
    7     8    1  0.47 2000
    8     9    1  0.47 2000
    9    10    1  0.47 2000
   10    11    1  0.47 2000
   11    12    1  0.47 2000
   12    13    1  0.47 2000
   13    14    1  0.47 2000
   14    15    1  0.47 2000
   15    16    1  0.47 2000
   16    17    1  0.47 2000
   17    18    1  0.47 2000
   18    19    1  0.47 2000
   19    20    1  0.47 2000
   20    21    1  0.47 2000
   21    22    1  0.47 2000
    [ moleculetype ]
    testB 1
    [ atoms ]
    1    N0   1   ASP    BB   1 0.00     45
    2    N0   2   GLY    SC1  1 0.00     45
    3    N0   3   ASP    SC2  1 0.00     45
    4    N0   4   ASP    BB   2 0.00     45
    5    N0   5   GLY    SC1  2 0.00     45
    6    N0   6   ASP    SC2  2 0.00     45
    7    N0   7   ASP    SC2  2 0.00     45
    8    N0   8   GLU    SC2  2 0.00     45
    9    N0   9   GLU    SC2  2 0.00     45
   10    N0  10   GLY    SC2  2 0.00     45
   11    N0  11   ASP    SC2  2 0.00     45
   12    N0  12   GLU    SC2  2 0.00     45
   13    N0  13   ASP    SC2  2 0.00     45
   14    N0  14   ASP    SC2  2 0.00     45
   15    N0  15   GLU    SC2  2 0.00     45
   16    N0  16   ASP    SC2  2 0.00     45
   17    N0  17   ASP    SC2  2 0.00     45
   18    N0  18   GLY    SC2  2 0.00     45
   19    N0  19   ASP    SC2  2 0.00     45
   20    N0  20   ASP    SC2  2 0.00     45
   21    N0  21   GLY    SC2  2 0.00     45
   22    N0  22   ASP    SC2  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    6     7    1  0.47 2000
    7     8    1  0.47 2000
    8     9    1  0.47 2000
    9    10    1  0.47 2000
   10    11    1  0.47 2000
   11    12    1  0.47 2000
   12    13    1  0.47 2000
   13    14    1  0.47 2000
   14    15    1  0.47 2000
   15    16    1  0.47 2000
   16    17    1  0.47 2000
   17    18    1  0.47 2000
   18    19    1  0.47 2000
   19    20    1  0.47 2000
   20    21    1  0.47 2000
   21    22    1  0.47 2000
    [ moleculetype ]
    testC 1
    [ atoms ]
    1    N0   1   GLY    BB   1 0.00     45
    2    N0   2   GLY    SC1  1 0.00     45
    3    N0   3   GLY    SC2  1 0.00     45
    4    N0   4   GLU    BB   2 0.00     45
    5    N0   5   GLU    SC1  2 0.00     45
    6    N0   6   GLU    SC2  2 0.00     45
    7    N0   7   GLU    SC2  2 0.00     45
    8    N0   8   GLU    SC2  2 0.00     45
    9    N0   9   GLU    SC2  2 0.00     45
   10    N0  10   GLU    SC2  2 0.00     45
   11    N0  11   GLU    SC2  2 0.00     45
   12    N0  12   GLU    SC2  2 0.00     45
   13    N0  13   GLU    SC2  2 0.00     45
   14    N0  14   GLU    SC2  2 0.00     45
   15    N0  15   GLU    SC2  2 0.00     45
   16    N0  16   GLU    SC2  2 0.00     45
   17    N0  17   GLU    SC2  2 0.00     45
   18    N0  18   GLU    SC2  2 0.00     45
   19    N0  19   GLU    SC2  2 0.00     45
   20    N0  20   GLU    SC2  2 0.00     45
   21    N0  21   GLU    SC2  2 0.00     45
   22    N0  22   GLU    SC2  2 0.00     45
   23    N0  23   GLU    SC2  2 0.00     45
   24    N0  24   GLU    SC2  2 0.00     45
   25    N0  25   GLU    SC2  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    6     7    1  0.47 2000
    7     8    1  0.47 2000
    8     9    1  0.47 2000
    9    10    1  0.47 2000
   10    11    1  0.47 2000
   11    12    1  0.47 2000
   12    13    1  0.47 2000
   13    14    1  0.47 2000
   14    15    1  0.47 2000
   15    16    1  0.47 2000
   16    17    1  0.47 2000
   17    18    1  0.47 2000
   18    19    1  0.47 2000
   19    20    1  0.47 2000
   20    21    1  0.47 2000
   21    22    1  0.47 2000
    3    23    1  0.47 2000
    23   24    1  0.47 2000
    3    25    1  0.47 2000
   [ system ]
    test system
    [ molecules ]
    testA 10
    testB 10
    testC 5
    """
    lines = textwrap.dedent(top_lines)
    lines = lines.splitlines()
    force_field = ForceField("test")
    topology = Topology(force_field)
    read_topology(lines=lines, topology=topology, cwdir="./")
    topology.preprocess()
    topology.volumes = {"GLY": 0.53, "GLU": 0.67, "ASP": 0.43}
    return topology

#PersistenceSpecs = namedtuple("presist", ["model", "lp", "start", "stop", "mol_idxs"])

@pytest.mark.parametrize('specs, seed, avg_step, expected', (
   # single restrain
   (
    [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(0, 10)))],
    63594,
    [0.6533],
    [5.88, 4.57333333, 5.88, 7.84, 7.84, 9.14666667, 7.84, 3.26666667, 5.22666667, 6.53333333],
   ),
   # single restrain branched
   (
    [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(20, 24)))],
    63594,
    [0.6533],
    [5.88, 4.57333333, 5.88, 7.84],
   ),
   # single restraint different random seed
   (
    [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(0, 10)))],
    23893,
    [0.6533],
    [6.53333333, 4.57333333, 9.14666667, 10.45333333, 6.53333333, 7.84, 7.84, 8.49333333, 7.84, 7.84],
   ),
   # smaller range; note parser requires consecutive mol_idxs
   (
    [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(0, 5)))],
    63594,
    [0.6533],
    [5.88, 4.57333333, 5.88, 7.84, 7.84],
   ),
   # test second group of molecules with mixed residues
   (
    [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(10, 15)))],
    63594,
    [0.4995],
    [5.4947619, 4.49571429, 5.4947619, 6.49380952, 6.49380952],
   ),
   # test two groups of the same molecule
   (
    [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(0, 5))),
     PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(5, 10))),],
    63594,
    [0.6533, 0.6533],
    [5.88, 4.57333333, 5.88, 7.84, 7.84, 5.88, 4.57333333, 5.88, 7.84, 7.84],
   ),
   # test three groups of the molecules two are the same one different
   (
    [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(0, 5))),
     PersistenceSpecs(model="WCM", lp=3.0, start=0, stop=21, mol_idxs=list(range(5, 10))),
     PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(10, 15))),],
    63594,
    [0.6533, 0.6533, 0.4995],
    [5.88, 4.573, 5.88, 7.84, 7.84, 9.1467, 8.4933, 9.146, 10.4533, 10.4533,
     5.498, 4.4957, 5.4947,  6.4938, 6.4938]
   )))
def test_persistence(topology, specs, seed, avg_step, expected):
    topology.persistences = specs
    nb_engine =  NonBondEngine.from_topology(topology.molecules,
                                             topology,
                                             box=np.array([15., 15., 15.]))

    sample_end_to_end_distances(topology, nb_engine, seed=seed)
    mol_count = 0
    for batch_count, batch in enumerate(specs):
        print(batch_count)
        for mol_idx in batch.mol_idxs:
            mol = topology.molecules[mol_idx]
            mol_copy = polyply.src.meta_molecule.MetaMolecule()
            mol_copy.add_edges_from(mol.edges)
            distance = expected[mol_count]
            avg_step_length = avg_step[batch_count]

            polyply.src.restraints.set_distance_restraint(mol_copy,
                                                          batch.stop,
                                                          batch.start,
                                                          distance,
                                                          avg_step_length,
                                                          tolerance=0.0)
            for node in mol.nodes:
                print(node)
                if "distance_restraints" in mol_copy.nodes[node]:
                    restr = mol.nodes[node]["distance_restraints"]
                    ref_restr = mol_copy.nodes[node]["distance_restraints"]
                    for new, ref in zip(restr, ref_restr):
                        assert new == pytest.approx(ref, rel=1e-3)
            mol_count += 1
        batch_count += 1

def test_error(topology):
    specs = [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(0, 10)))]
    seed = 63594
    avg_step = 0.6533
    topology.persistences = specs
    nb_engine = NonBondEngine.from_topology(topology.molecules,
                                            topology,
                                            box=np.array([1., 1., 1.]))
    with pytest.raises(IOError):
        sample_end_to_end_distances(topology, nb_engine, seed=seed)

def test_warning(caplog, topology):
    caplog.set_level(logging.WARNING)
    specs = [PersistenceSpecs(model="WCM", lp=1.5, start=0, stop=21, mol_idxs=list(range(0, 10)))]
    seed = 63594
    avg_step = 0.6533
    topology.persistences = specs
    nb_engine = NonBondEngine.from_topology(topology.molecules,
                                            topology,
                                            box=np.array([9.0, 9., 9.]))
    with caplog.at_level(logging.WARNING):
        sample_end_to_end_distances(topology, nb_engine, seed=seed)
        for record in caplog.records:
            assert record.levelname == "WARNING"
            break
        else:
            assert False



import logging
from vermouth.log_helpers import (StyleAdapter, BipolarFormatter,
                                  CountingHandler, TypeAdapter,)

# Implement Logger
LOGGER = TypeAdapter(logging.getLogger('polyply'))
PRETTY_FORMATTER = logging.Formatter(fmt='{levelname:} - {type} - {message}',
                                     style='{')
DETAILED_FORMATTER = logging.Formatter(fmt='{levelname:} - {type} - {name} - {message}',
                                       style='{')
COUNTER = CountingHandler()

# Control above what level message we want to count
COUNTER.setLevel(logging.WARNING)

CONSOLE_HANDLER = logging.StreamHandler()
FORMATTER = BipolarFormatter(DETAILED_FORMATTER,
                             PRETTY_FORMATTER,
                             logging.DEBUG,
                             logger=LOGGER)

CONSOLE_HANDLER.setFormatter(FORMATTER)
LOGGER.addHandler(CONSOLE_HANDLER)
LOGGER.addHandler(COUNTER)

LOGGER = StyleAdapter(LOGGER)

LOGLEVELS = {0: logging.INFO, 1: logging.DEBUG, 2: 5}



import pytest
import logging
import numpy as np
import polyply
from vermouth.log_helpers import (StyleAdapter, BipolarFormatter,
                                  CountingHandler, TypeAdapter,)
from polyply.tests.test_build_file_parser import test_molecule, test_system
from polyply.src.logging import LOGGER, LOGLEVELS

LOGGER.setLevel(LOGLEVELS[1])


@pytest.mark.parametrize('_type, option, expected, warning, idxs', (
   # tag all nodes 1-4 that have name ALA, GLU, THR should not get tagged and raise warning
   ("cylinder",
    {"resname": "ALA", "start": 1, "stop": 9, "parameters":["in", np.array([5.0, 5.0, 5.0]), 5.0, 5.0]},
    [0, 1, 2, 3],
    "parsing build file: could not find resid {} with resname ALA in molecule AA.",
    range(5, 9),
    ),
   # raise warning that residue random cannot be found
   ("sphere",
    {"resname": "ALA", "start": 9, "stop": 12, "parameters":["in", np.array([10.0, 10.0, 10.0]), 5.0]},
    [],
    "parsing build file: could not find resid {} with resname ALA in molecule AA.",
    range(9, 12)),
   ))
def test_tag_nodes_logging(caplog, test_molecule, _type, option, expected, warning, idxs):
    polyply.src.build_file_parser.BuildDirector._tag_nodes(test_molecule, _type, option, "AA")
    for record, idx in zip(caplog.records, idxs):
        assert record.getMessage() == warning.format(idx)
    for node in test_molecule.nodes:
        if "restraints" in test_molecule.nodes[node]:
           assert node in expected

def test_mol_directive_logging(caplog, test_system):
    processor = polyply.src.build_file_parser.BuildDirector([], test_system)
    line = "AA 0 3"
    processor._molecule(line)
    assert caplog.records[0].getMessage() == "parsing build file: could not find molecule with name AA and index 2."



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import networkx as nx
from vermouth.log_helpers import StyleAdapter, get_logger
from .graph_utils import compute_avg_step_length, get_all_predecessors
from .restraints import set_distance_restraint

LOGGER = StyleAdapter(get_logger(__name__))

def worm_like_chain_model(ee_dist, r_max, persistence_length):
    """
    Probability to find an end-to-end distance h given a contour length
    L and persistence length _lambda according to the Worm-Like-Chain
    model. The equation was published in: "Lee, H. and Pastor, R.W., 2011.
    Coarse-grained model for PEGylated lipids: effect of PEGylation on the
    size and shape of self-assembled structures. The Journal of Physical
    Chemistry B, 115(24), pp.7830-7837."

    Parameters
    ----------
    ee_dist: float
        the end-to-end distance
    r_max: float
        the contour length
    _lambda: float
        the persistence length

    Returns
    -------
    float
        probability to find the end-to-end distance
    """
    alpha = 3*r_max/(4* persistence_length)
    factor_1 = (np.pi**3/2. * np.exp(-alpha)*alpha**(-3/2.)*(1 + 3/(alpha) + (15/4)/alpha**2.))**-1.
    nominator = 4*np.pi*ee_dist**2.* factor_1

    denominator = r_max*(1-(ee_dist/r_max)**2.)**9/2.
    nominator_exp = -3*r_max
    denominator_exp = 4* persistence_length*(1-(ee_dist/r_max)**2.)

    return nominator/denominator * np.exp(nominator_exp/denominator_exp)


DISTRIBUTIONS = {"WCM": worm_like_chain_model, }

def generate_end_end_distances(specs,
                               avg_step_length,
                               max_path_length,
                               box,
                               limit_prob=1e-5,
                               seed=None):
    """
    Subsample a distribution of end-to-end distances given a
    persistence length, residue graph, and theoretical model.

    Parameters
    ----------
    specs: `tuple`
        named tuple with attributes model, lp (i.e. persistence length),
        start, stop node indices, which define the ends and mol_idxs
        which are the indices of the batch of molecules.
    avg_step_length: float
        average step length (nm)
    max_path_length: float
        total length of the path (nm)
    box: np.ndarray
        box vectors
    limit_prob: float
        the lowest probability after which to truncate end-to-end distance
        distribution
    seed: int
        random seed for subsampling distribution

    Returns
    -------
    np.ndarray
        an array of end-to-end distances with shape (1, len(mol_idxs))
    """
    ee_distances = np.arange(avg_step_length, max_path_length, avg_step_length)
    # generate the distribution and sample it
    probabilities = DISTRIBUTIONS[specs.model](ee_distances, max_path_length, specs.lp)
    # filter the end-to-end distances
    ee_distances = ee_distances[probabilities > limit_prob]
    probabilities = probabilities[probabilities > limit_prob]
    #TODO
    #replace choice by https://numpy.org/doc/stable/reference/random/generated/
    #                          numpy.random.Generator.choice.html#numpy.random.Generator.choice
    np.random.seed(seed)
    ee_samples = np.random.choice(ee_distances,
                                  p=probabilities/np.sum(probabilities),
                                  size=len(specs.mol_idxs))
    # raise warning if the smallest box vector is smaller than the largest end-to-end
    # distance
    if max(ee_samples) > np.sqrt(3*min(box)**2):
        msg = ("Sampling the end-to-end distance distribution yielded end-to-end distances\n"
               "that are larger than the box diagonal. This will not work. Please increase the\n"
               "box to be at least {} nm.")
        raise IOError(msg.format(max(ee_samples)))

    if max(ee_samples) > min(box):
        msg = ("Your smallest box vector {} is smaller than the largest end-to-end distance {},\n"
               "which is dictated by the set persistence length. This can prevent the algorithm \n"
               "from converging, since it will not be able to place the molecule in the \n"
               "relatively small box dimensions. You should increase the boxsize to be at \n"
               "least {} ideally even a bit larger.")
        # add a 10% buffer just to be sure
        LOGGER.warning(msg, min(box), max(ee_samples), max(ee_samples)*1.1)

    return ee_samples

def sample_end_to_end_distances(topology, nonbond_matrix, seed=None):
    """
    Apply distance restraints to the ends of molecules given a distribution
    of end-to-end distances generated from a given theoreical model. The
    topology attribute persistences contains a list of batches of molecules
    for which a persistence length has been given. This function generates
    the corresponding restraints taking into account the actual path
    lengths as defined by the volumes in the super-CG model.

    Parameters
    ----------
    topology: `class:polyply.src.topology.Topology`
        the topology of the system
    nonbond_matrix: `:class:polyply.src.nb_engine.NonBondMatrix`
        the nonbonded matrix that stores the interactions of the
        super CG model built from the topology
    seed:
        random seed for subsampling distribution

    Returns
    -------
    list[`:class:nx.Graph`]
        list of updated molecules
    """
    # loop over all batches of molecules
    for specs in topology.persistences:
        molecule = topology.molecules[specs.mol_idxs[0]]

        # first we initalize the path
        molecule.root = specs.start
        path = get_all_predecessors(molecule.search_tree,
                                    start_node=specs.start,
                                    node=specs.stop)
        edge_path = list(zip(path[:-1], path[1:]))
        # compute the average step length end-to-end
        avg_step_length, max_length = compute_avg_step_length(molecule,
                                                              specs.mol_idxs[0],
                                                              nonbond_matrix,
                                                              edge_path)
        # generate a distribution of end-to-end distances
        distribution = generate_end_end_distances(specs,
                                                  avg_step_length,
                                                  max_length,
                                                  box=nonbond_matrix.boxsize,
                                                  seed=seed)

        # for each molecule in the batch assign one end-to-end
        # distance restraint from the distribution.
        for mol_idx, dist in zip(specs.mol_idxs, distribution):
            # we set the bfs root here to enforce that start
            # is placed before stop
            topology.molecules[mol_idx].root = specs.start
            set_distance_restraint(topology.molecules[mol_idx],
                                   specs.stop,
                                   specs.start,
                                   dist,
                                   avg_step_length,
                                   tolerance=0.0)



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that inequivalent residues are identified.
"""
import textwrap
import pytest
from vermouth.forcefield import ForceField
import polyply
from polyply.src.topology import Topology
from polyply.src.top_parser import read_topology

@pytest.mark.parametrize('top_lines', (
    # residue in two different molecules not equivalent by
    # number of atoms
    ("""
    [ defaults ]
    1   1   no   1.0     1.0
    [ atomtypes ]
    N0 72.0 0.000 A 0.0 0.0
    N1 72.0 0.000 A 0.0 0.0
    [ nonbond_params ]
    N0   N0   1 4.700000e-01    3.700000e+00
    N1   N1   1 4.700000e-01    3.700000e+00
    [ moleculetype ]
    testA 1
    [ atoms ]
    1    N0   1   GLY    BB   1 0.00     45
    2    N0   1   GLY    SC1  1 0.00     45
    3    N0   1   GLY    SC2  1 0.00     45
    4    N1   2   GLU    BB   2 0.00     45
    5    N1   2   GLU    SC1  2 0.00     45
    6    N1   2   GLU    SC2  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    [ moleculetype ]
    testB 1
    [ atoms ]
    1    N0   1   ASP    BB   1 0.00     45
    2    N0   1   ASP    SC3  1 0.00     45
    3    N0   1   ASP    SC4  1 0.00     45
    4    N1   2   GLU    BB   2 0.00     45
    5    N1   2   GLU    SC1  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    [ system ]
    test system
    [ molecules ]
    testA 1
    testB 1
    """),
    # residue in two different molecules not equivalent by
    # connectivity
    ("""
    [ defaults ]
    1   1   no   1.0     1.0
    [ atomtypes ]
    N0 72.0 0.000 A 0.0 0.0
    N1 72.0 0.000 A 0.0 0.0
    [ nonbond_params ]
    N0   N0   1 4.700000e-01    3.700000e+00
    N1   N1   1 4.700000e-01    3.700000e+00
    [ moleculetype ]
    testA 1
    [ atoms ]
    1    N0   1   GLY    BB   1 0.00     45
    2    N0   1   GLY    SC1  1 0.00     45
    3    N0   1   GLY    SC2  1 0.00     45
    4    N1   2   GLU    BB   2 0.00     45
    5    N1   2   GLU    SC1  2 0.00     45
    6    N1   2   GLU    SC2  2 0.00     45
    7    N1   2   GLU    SC3  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    6     7    1  0.47 2000
    [ moleculetype ]
    testB 1
    [ atoms ]
    1    N0   1   ASP    BB   1 0.00     45
    2    N0   1   ASP    SC3  1 0.00     45
    3    N0   1   ASP    SC4  1 0.00     45
    4    N1   2   GLU    BB   2 0.00     45
    5    N1   2   GLU    SC1  2 0.00     45
    6    N1   2   GLU    SC2  2 0.00     45
    7    N1   2   GLU    SC3  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    5     7    1  0.47 2000
    [ system ]
    test system
    [ molecules ]
    testA 1
    testB 1
    """)))
def test_raise_residue_error(top_lines):
    lines = textwrap.dedent(top_lines)
    lines = lines.splitlines()
    force_field = ForceField("test")
    topology = Topology(force_field)
    read_topology(lines=lines, topology=topology, cwdir="./")
    topology.preprocess()
    topology.volumes = {"GLY": 0.53, "GLU": 0.67, "ASP": 0.43}
    with pytest.raises(IOError):
        polyply.src.check_residue_equivalence.check_residue_equivalence(topology)

def test_raise_residue_no_error():
    """
    This test makes sure that we actually skip molecules that are defined
    in the top file but not used in the actual system.
    """
    top_lines="""
    [ defaults ]
    1   1   no   1.0     1.0
    [ atomtypes ]
    N0 72.0 0.000 A 0.0 0.0
    N1 72.0 0.000 A 0.0 0.0
    [ nonbond_params ]
    N0   N0   1 4.700000e-01    3.700000e+00
    N1   N1   1 4.700000e-01    3.700000e+00
    [ moleculetype ]
    testA 1
    [ atoms ]
    1    N0   1   GLY    BB   1 0.00     45
    2    N0   1   GLY    SC1  1 0.00     45
    3    N0   1   GLY    SC2  1 0.00     45
    4    N1   2   GLU    BB   2 0.00     45
    5    N1   2   GLU    SC1  2 0.00     45
    6    N1   2   GLU    SC2  2 0.00     45
    7    N1   2   GLU    SC3  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    6     7    1  0.47 2000
    [ moleculetype ]
    testB 1
    [ atoms ]
    1    N0   1   ASP    BB   1 0.00     45
    2    N0   1   ASP    SC3  1 0.00     45
    3    N0   1   ASP    SC4  1 0.00     45
    4    N1   2   GLU    BB   2 0.00     45
    5    N1   2   GLU    SC1  2 0.00     45
    6    N1   2   GLU    SC2  2 0.00     45
    7    N1   2   GLU    SC3  2 0.00     45
    [ bonds ]
    1     2    1  0.47 2000
    2     3    1  0.47 2000
    3     4    1  0.47 2000
    4     5    1  0.47 2000
    5     6    1  0.47 2000
    5     7    1  0.47 2000
    [ system ]
    test system
    [ molecules ]
    testA 1
    """
    lines = textwrap.dedent(top_lines)
    lines = lines.splitlines()
    force_field = ForceField("test")
    topology = Topology(force_field)
    read_topology(lines=lines, topology=topology, cwdir="./")
    topology.preprocess()
    topology.volumes = {"GLY": 0.53, "GLU": 0.67, "ASP": 0.43}
    polyply.src.check_residue_equivalence.check_residue_equivalence(topology)
